
//////////////////////////////////////////////////////////////
//
//          FMotifs
//                      A program for the fast evaluation of
//                      motifs in dense medium-size networks.
//                      Developed by M. Zanin.
//
//                      V2.0
//
//                      Check www.mzanin.com/FMotifs for further
//                      details and usage examples.
//
//                      Download the manual at:
//                      http://www.mzanin.com/FMotifs/FMotifs%20Guide.pdf
//


#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <math.h>

#include "FastMotifs3.h"


//  ------------------------------------------------------------
//  Class constructor

FastMotifs3::FastMotifs3()
{
    for(int i = 0; i < 13; ++ i)
        Motifs[i] = 0;
}


//  ------------------------------------------------------------
//  Class destruction

FastMotifs3::~FastMotifs3()
{
}



//  ------------------------------------------------------------
//  Returns the version of the class

float FastMotifs3::GetVersion()
{
    return 2.0;
}



//  ------------------------------------------------------------
//  Allocates memory for the adjacency matrix describing the network
//  Input arguments:        nN      Number of nodes

void FastMotifs3::AllocMemory(int nN)
{
    numNodes = nN;
    AM = (bool *) malloc( sizeof(bool) * nN * nN );

    for( int i = 0; i < nN * nN; ++ i )
        AM[ i ] = false;

    IsConnected = (bool *) malloc( sizeof(bool) * nN );
}



//  ------------------------------------------------------------
//  Clears the memory used to store the adjacency matrix.
//  To be used when different networks should be analyzed with
//  the same class.

void FastMotifs3::ClearMemory()
{
    if( AM != 0 )
        free( AM );

    if( IsConnected != 0 )
        free( IsConnected );
}



//  ------------------------------------------------------------
//  Main motifs calculation function.
//  Takes as input the adjacency matrix (as stored in AM), and returns
//  the motif frequencies in the array Motifs.

void FastMotifs3::CalculateMotifs()
{
int node1, node2, node3;


    for(int i = 0; i < 512; ++ i)
        MotifsArray[i] = 0;
    CalculateConnections();


    for( node1 = 0; node1 < numNodes; ++ node1 ){

        if( ! IsConnected[node1] ) continue;

        for( node2 = node1 + 1; node2 < numNodes; ++ node2 ){

            if( ! IsConnected[node2] ) continue;

            int XReg = 0;

            if( AM[ node1 * numNodes + node2 ] )
                XReg = SetBit(XReg, 1);
            if( AM[ node2 * numNodes + node1 ] )
                XReg = SetBit(XReg, 3);

            for( node3 = node2 + 1; node3 < numNodes; ++ node3 ){

                if( ! IsConnected[node3] ) continue;

                int Reg = XReg;
                if( AM[ node1 * numNodes + node3 ] )
                    Reg = SetBit(Reg, 2);
                if( AM[ node2 * numNodes + node3 ] )
                    Reg = SetBit(Reg, 5);
                if( AM[ node3 * numNodes + node1 ] )
                    Reg = SetBit(Reg, 6);
                if( AM[ node3 * numNodes + node2 ] )
                    Reg = SetBit(Reg, 7);

                ++ MotifsArray[ Reg ];

            }
        }
    }


    DecryptMotifs();

}



//  ------------------------------------------------------------
//  Pre-calculates the nodes that should be analyzed by each core
//  in parallel mode.

void FastMotifs3::ParCreateIndices(int numCores)
{
    ParIndices = (int *) malloc( sizeof(int) * (numCores + 1) );

    double iter = ParNumIterations() / numCores;
    ParIndices[0] = 0;

    for(int i = 0; i < numCores; i ++){
        int t = ParTargetNode(ParIndices[i], iter);
        ParIndices[i + 1] = t;
    }
    ParIndices[ numCores ] = numNodes;

    for(int i = 0; i < numCores + 1; i ++)
        ParIndices[i] = numNodes - ParIndices[i];

}



//  ------------------------------------------------------------
//  Main motifs calculation function in parallel mode.
//  Takes as input the adjacency matrix (as stored in AM), and returns
//  the motif frequencies in the array Motifs.
//  Only the nodes associated with the selected core are processed.

void FastMotifs3::ParCalculateMotifs(int idCore)
{
    int node1, node2, node3;


    CalculateConnections();
    for(int i = 0; i < 512; ++ i)
        MotifsArray[i] = 0;


    for( node1 = ParIndices[idCore+1]; node1 < ParIndices[idCore]; ++ node1 ){

        if( ! IsConnected[node1] ) continue;

        for( node2 = node1 + 1; node2 < numNodes; ++ node2 ){

            if( ! IsConnected[node2] ) continue;

            int XReg = 0;

            if( AM[ node1 * numNodes + node2 ] )
                XReg = SetBit(XReg, 1);
            if( AM[ node2 * numNodes + node1 ] )
                XReg = SetBit(XReg, 3);

            for( node3 = node2 + 1; node3 < numNodes; ++ node3 ){

                if( ! IsConnected[node3] ) continue;

                int Reg = XReg;
                if( AM[ node1 * numNodes + node3 ] )
                    Reg = SetBit(Reg, 2);
                if( AM[ node2 * numNodes + node3 ] )
                    Reg = SetBit(Reg, 5);
                if( AM[ node3 * numNodes + node1 ] )
                    Reg = SetBit(Reg, 6);
                if( AM[ node3 * numNodes + node2 ] )
                    Reg = SetBit(Reg, 7);

                ++ MotifsArray[ Reg ];

            }
        }
    }


    DecryptMotifs();

}




//  ------------------------------------------------------------
//                      INTERNAL FUNCTION
//  ------------------------------------------------------------



//  ------------------------------------------------------------
//  Converts the motif hits stored in the array MotifsArray in
//  motif frequencies (stored in the array Motifs).

void FastMotifs3::DecryptMotifs()
{
int i;
int Offset;

    for(i = 0; i < 13; ++ i)
        Motifs[i] = 0;


    //  ---------------------------
    //      Motif 1

    Offset = SetBit(0, 1) + SetBit(0, 2);
    Motifs[0] += MotifsArray[ Offset ];

    Offset = SetBit(0, 3) + SetBit(0, 5);
    Motifs[0] += MotifsArray[ Offset ];

    Offset = SetBit(0, 6) + SetBit(0, 7);
    Motifs[0] += MotifsArray[ Offset ];


    //  ---------------------------
    //      Motif 2

    Offset = SetBit(0, 2) + SetBit(0, 3);
    Motifs[1] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 6);
    Motifs[1] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 5);
    Motifs[1] += MotifsArray[ Offset ];

    Offset = SetBit(0, 3) + SetBit(0, 7);
    Motifs[1] += MotifsArray[ Offset ];

    Offset = SetBit(0, 2) + SetBit(0, 7);
    Motifs[1] += MotifsArray[ Offset ];

    Offset = SetBit(0, 5) + SetBit(0, 6);
    Motifs[1] += MotifsArray[ Offset ];


    //  ---------------------------
    //      Motif 3

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 3);
    Motifs[2] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 6);
    Motifs[2] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 3) + SetBit(0, 5);
    Motifs[2] += MotifsArray[ Offset ];

    Offset = SetBit(0, 3) + SetBit(0, 5) + SetBit(0, 7);
    Motifs[2] += MotifsArray[ Offset ];

    Offset = SetBit(0, 2) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[2] += MotifsArray[ Offset ];

    Offset = SetBit(0, 5) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[2] += MotifsArray[ Offset ];


    //  ---------------------------
    //      Motif 4

    Offset = SetBit(0, 2) + SetBit(0, 5);
    Motifs[3] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 7);
    Motifs[3] += MotifsArray[ Offset ];

    Offset = SetBit(0, 3) + SetBit(0, 6);
    Motifs[3] += MotifsArray[ Offset ];


    //  ---------------------------
    //      Motif 5

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 5);
    Motifs[4] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 7);
    Motifs[4] += MotifsArray[ Offset ];

    Offset = SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 5);
    Motifs[4] += MotifsArray[ Offset ];

    Offset = SetBit(0, 3) + SetBit(0, 5) + SetBit(0, 6);
    Motifs[4] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[4] += MotifsArray[ Offset ];

    Offset = SetBit(0, 3) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[4] += MotifsArray[ Offset ];



    //  ---------------------------
    //      Motif 6

    Offset = SetBit(0, 3) + SetBit(0, 5) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[5] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[5] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 5);
    Motifs[5] += MotifsArray[ Offset ];


    //  ---------------------------
    //      Motif 7

    Offset = SetBit(0, 1) + SetBit(0, 3) + SetBit(0, 6);
    Motifs[6] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 3) + SetBit(0, 7);
    Motifs[6] += MotifsArray[ Offset ];

    Offset = SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 6);
    Motifs[6] += MotifsArray[ Offset ];

    Offset = SetBit(0, 2) + SetBit(0, 5) + SetBit(0, 6);
    Motifs[6] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 5) + SetBit(0, 7);
    Motifs[6] += MotifsArray[ Offset ];

    Offset = SetBit(0, 2) + SetBit(0, 5) + SetBit(0, 7);
    Motifs[6] += MotifsArray[ Offset ];



    //  ---------------------------
    //      Motif 8

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 6);
    Motifs[7] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 3) + SetBit(0, 5) + SetBit(0, 7);
    Motifs[7] += MotifsArray[ Offset ];

    Offset = SetBit(0, 2) + SetBit(0, 5) + SetBit(0, 7) + SetBit(0, 6);
    Motifs[7] += MotifsArray[ Offset ];


    //  ---------------------------
    //      Motif 9

    Offset = SetBit(0, 1) + SetBit(0, 5) + SetBit(0, 6);
    Motifs[8] += MotifsArray[ Offset ];

    Offset = SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 7);
    Motifs[8] += MotifsArray[ Offset ];



    //  ---------------------------
    //      Motif 10

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 5) + SetBit(0, 6);
    Motifs[9] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 3) + SetBit(0, 5) + SetBit(0, 6);
    Motifs[9] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 5) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[9] += MotifsArray[ Offset ];

    Offset = SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[9] += MotifsArray[ Offset ];

    Offset = SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 5) + SetBit(0, 7);
    Motifs[9] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 7);
    Motifs[9] += MotifsArray[ Offset ];



    //  ---------------------------
    //      Motif 11

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 5) + SetBit(0, 7);
    Motifs[10] += MotifsArray[ Offset ];

    Offset = SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 5) + SetBit(0, 6);
    Motifs[10] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 3) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[10] += MotifsArray[ Offset ];



    //  ---------------------------
    //      Motif 12

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 5) + SetBit(0, 6);
    Motifs[11] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[11] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 3) + SetBit(0, 5) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[11] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 5) + SetBit(0, 7);
    Motifs[11] += MotifsArray[ Offset ];

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 5) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[11] += MotifsArray[ Offset ];

    Offset = SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 5) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[11] += MotifsArray[ Offset ];



    //  ---------------------------
    //      Motif 13

    Offset = SetBit(0, 1) + SetBit(0, 2) + SetBit(0, 3) + SetBit(0, 5) + SetBit(0, 6) + SetBit(0, 7);
    Motifs[12] += MotifsArray[ Offset ];




}



//  ------------------------------------------------------------
//  Sets to 1 a given bit inside the variable Value

inline int FastMotifs3::SetBit(int Value, int Offset)
{
    return Value | ( 1 << ( 8 - Offset ) );
}



//  ------------------------------------------------------------
//  Sets to 0 a given bit inside the variable Value

inline int FastMotifs3::ResetBit(int Value, int Offset)
{
    return Value & ( ~ ( 1 << ( 8 - Offset) ) );
}



//  ------------------------------------------------------------
//  Pre-identifies the nodes that are not connected.
//  Used to improve calculation performance on very sparse networks.

void FastMotifs3::CalculateConnections()
{
register int i, t;

    for( t = 0; t < numNodes; ++ t ){
        IsConnected[t] = false;
        for( i = 0; i < numNodes; ++ i )
            if( AM[t * numNodes + i] || AM[i * numNodes + t] ){
                IsConnected[t] = true;
                break;
            }
    }
}



//  ------------------------------------------------------------
//  Calculates the number of iterations needed to analyze the graph.
//  It is used to balance the computational time across different cores.

double FastMotifs3::ParNumIterations()
{
    return ( pow((double)this->numNodes, 3) - (double)this->numNodes ) / 6.0;
}



//  ------------------------------------------------------------
//  Calculates the final node that should be analyzed by a given core.
//  It is used to balance the computational time across different cores.

int FastMotifs3::ParTargetNode(int inNode, double numIterations)
{
    double g = pow(inNode, 3.0) - 3.0 * inNode * inNode + 2.0 * inNode + 6.0 * numIterations;
    double x = CubeRoot( g );

    return floor(x);
}



//  ------------------------------------------------------------
//  Fast calculation of the cube root.

inline float FastMotifs3::CubeRoot(float X)
{
    if (X < 0)
        return -1.0 * pow(-X, 1.0 / 3.0);
    else
        return pow(X, 1.0 / 3.0);
}




